import typing as t
from dataclasses import dataclass
from types import MappingProxyType


@dataclass(frozen=True)
class Language:
    """Language used in a post."""

    name: str  #: English name of Language.
    code1: str  #: ISO 639-1 Code.
    code2: str  #: ISO 639-2 Code.


# language databases:
# - official db: https://www.loc.gov/standards/iso639-2/php/code_list.php
# - bsky app: https://github.com/bluesky-social/social-app/blob/main/src/locale/languages.ts
LANGUAGES: t.Tuple = (
    Language(name='Afar', code1='aa', code2='aar'),
    Language(name='Abkhazian', code1='ab', code2='abk'),
    Language(name='Achinese', code1='', code2='ace'),
    Language(name='Acoli', code1='', code2='ach'),
    Language(name='Adangme', code1='', code2='ada'),
    Language(name='Adyghe; Adygei', code1='', code2='ady'),
    Language(name='Afro-Asiatic languages', code1='', code2='afa'),
    Language(name='Afrihili', code1='', code2='afh'),
    Language(name='Afrikaans', code1='af', code2='afr'),
    Language(name='Ainu', code1='', code2='ain'),
    Language(name='Akan', code1='ak', code2='aka'),
    Language(name='Akkadian', code1='', code2='akk'),
    Language(name='Albanian', code1='sq', code2='alb'),
    Language(name='Aleut', code1='', code2='ale'),
    Language(name='Algonquian languages', code1='', code2='alg'),
    Language(name='Southern Altai', code1='', code2='alt'),
    Language(name='Amharic', code1='am', code2='amh'),
    Language(name='English, Old (ca.450-1100)', code1='', code2='ang'),
    Language(name='Angika', code1='Angika', code2='anp '),
    Language(name='Apache languages', code1='', code2='apa'),
    Language(name='Arabic', code1='ar', code2='ara'),
    Language(name='Official Aramaic (700-300 BCE); Imperial Aramaic (700-300 BCE)', code1='', code2='arc'),
    Language(name='Aragonese', code1='an', code2='arg'),
    Language(name='Armenian', code1='hy', code2='arm'),
    Language(name='Mapudungun; Mapuche', code1='', code2='arn'),
    Language(name='Arapaho', code1='', code2='arp'),
    Language(name='Artificial languages', code1='', code2='art'),
    Language(name='Arawak', code1='', code2='arw'),
    Language(name='Assamese', code1='as', code2='asm'),
    Language(name='Asturian; Bable; Leonese; Asturleonese', code1='', code2='ast'),
    Language(name='Athapascan languages', code1='', code2='ath'),
    Language(name='Australian languages', code1='', code2='aus'),
    Language(name='Avaric', code1='av', code2='ava'),
    Language(name='Avestan', code1='ae', code2='ave'),
    Language(name='Awadhi', code1='', code2='awa'),
    Language(name='Aymara', code1='ay', code2='aym'),
    Language(name='Azerbaijani', code1='az', code2='aze'),
    Language(name='Banda languages', code1='', code2='bad'),
    Language(name='Bamileke languages', code1='', code2='bai'),
    Language(name='Bashkir', code1='ba', code2='bak'),
    Language(name='Baluchi', code1='', code2='bal'),
    Language(name='Bambara', code1='bm', code2='bam'),
    Language(name='Balinese', code1='', code2='ban'),
    Language(name='Basque', code1='eu', code2='baq'),
    Language(name='Basa', code1='', code2='bas'),
    Language(name='Baltic languages', code1='', code2='bat'),
    Language(name='Beja; Bedawiyet', code1='', code2='bej'),
    Language(name='Belarusian', code1='be', code2='bel'),
    Language(name='Bemba', code1='', code2='bem'),
    Language(name='Bengali', code1='bn', code2='ben'),
    Language(name='Berber languages', code1='', code2='ber'),
    Language(name='Bhojpuri', code1='', code2='bho'),
    Language(name='Bihari languages', code1='bh', code2='bih'),
    Language(name='Bikol', code1='', code2='bik'),
    Language(name='Bini; Edo', code1='', code2='bin'),
    Language(name='Bislama', code1='bi', code2='bis'),
    Language(name='Siksika', code1='', code2='bla'),
    Language(name='Bantu languages', code1='', code2='bnt'),
    Language(name='Tibetan', code1='bo', code2='bod'),
    Language(name='Bosnian', code1='bs', code2='bos'),
    Language(name='Braj', code1='', code2='bra'),
    Language(name='Breton', code1='br', code2='bre'),
    Language(name='Batak languages', code1='', code2='btk'),
    Language(name='Buriat', code1='', code2='bua'),
    Language(name='Buginese', code1='', code2='bug'),
    Language(name='Bulgarian', code1='bg', code2='bul'),
    Language(name='Burmese', code1='my', code2='bur'),
    Language(name='Blin; Bilin', code1='', code2='byn'),
    Language(name='Caddo', code1='', code2='cad'),
    Language(name='Central American Indian languages', code1='', code2='cai'),
    Language(name='Galibi Carib', code1='', code2='car'),
    Language(name='Catalan; Valencian', code1='ca', code2='cat'),
    Language(name='Caucasian languages', code1='', code2='cau'),
    Language(name='Cebuano', code1='', code2='ceb'),
    Language(name='Celtic languages', code1='', code2='cel'),
    Language(name='Czech', code1='cs', code2='ces'),
    Language(name='Chamorro', code1='ch', code2='cha'),
    Language(name='Chibcha', code1='', code2='chb'),
    Language(name='Chechen', code1='ce', code2='che'),
    Language(name='Chagatai', code1='', code2='chg'),
    Language(name='Chinese', code1='zh', code2='chi'),
    Language(name='Chuukese', code1='', code2='chk'),
    Language(name='Mari', code1='', code2='chm'),
    Language(name='Chinook jargon', code1='', code2='chn'),
    Language(name='Choctaw', code1='', code2='cho'),
    Language(name='Chipewyan; Dene Suline', code1='', code2='chp'),
    Language(name='Cherokee', code1='', code2='chr'),
    Language(
        name='Church Slavic; Old Slavonic; Church Slavonic; Old Bulgarian; Old Church Slavonic', code1='cu', code2='chu'
    ),
    Language(name='Chuvash', code1='cv', code2='chv'),
    Language(name='Cheyenne', code1='', code2='chy'),
    Language(name='Chamic languages', code1='', code2='cmc'),
    Language(name='Montenegrin', code1='', code2='cnr'),
    Language(name='Coptic', code1='', code2='cop'),
    Language(name='Cornish', code1='kw', code2='cor'),
    Language(name='Corsican', code1='co', code2='cos'),
    Language(name='Creoles and pidgins, English based', code1='', code2='cpe'),
    Language(name='Creoles and pidgins, French-based', code1='', code2='cpf'),
    Language(name='Creoles and pidgins, Portuguese-based', code1='', code2='cpp'),
    Language(name='Cree', code1='cr', code2='cre'),
    Language(name='Crimean Tatar; Crimean Turkish', code1='', code2='crh'),
    Language(name='Creoles and pidgins', code1='', code2='crp'),
    Language(name='Kashubian', code1='', code2='csb'),
    Language(name='Cushitic languages', code1='', code2='cus'),
    Language(name='Welsh', code1='cy', code2='cym'),
    Language(name='Czech', code1='cs', code2='cze'),
    Language(name='Dakota', code1='', code2='dak'),
    Language(name='Danish', code1='da', code2='dan'),
    Language(name='Dargwa', code1='', code2='dar'),
    Language(name='Land Dayak languages', code1='', code2='day'),
    Language(name='Delaware', code1='', code2='del'),
    Language(name='Slave (Athapascan)', code1='', code2='den'),
    Language(name='German', code1='de', code2='deu'),
    Language(name='Dogrib', code1='', code2='dgr'),
    Language(name='Dinka', code1='', code2='din'),
    Language(name='Divehi; Dhivehi; Maldivian', code1='dv', code2='div'),
    Language(name='Dogri', code1='', code2='doi'),
    Language(name='Dravidian languages', code1='', code2='dra'),
    Language(name='Lower Sorbian', code1='', code2='dsb'),
    Language(name='Duala', code1='', code2='dua'),
    Language(name='Dutch, Middle (ca.1050-1350)', code1='', code2='dum'),
    Language(name='Dutch; Flemish', code1='nl', code2='dut'),
    Language(name='Dyula', code1='', code2='dyu'),
    Language(name='Dzongkha', code1='dz', code2='dzo'),
    Language(name='Efik', code1='', code2='efi'),
    Language(name='Egyptian (Ancient)', code1='', code2='egy'),
    Language(name='Ekajuk', code1='', code2='eka'),
    Language(name='Greek, Modern (1453-)', code1='el', code2='ell'),
    Language(name='Elamite', code1='', code2='elx'),
    Language(name='English', code1='en', code2='eng'),
    Language(name='English, Middle (1100-1500)', code1='', code2='enm'),
    Language(name='Esperanto', code1='eo', code2='epo'),
    Language(name='Estonian', code1='et', code2='est'),
    Language(name='Basque', code1='eu', code2='eus'),
    Language(name='Ewe', code1='ee', code2='ewe'),
    Language(name='Ewondo', code1='', code2='ewo'),
    Language(name='Fang', code1='', code2='fan'),
    Language(name='Faroese', code1='fo', code2='fao'),
    Language(name='Persian', code1='fa', code2='fas'),
    Language(name='Fanti', code1='', code2='fat'),
    Language(name='Fijian', code1='fj', code2='fij'),
    Language(name='Filipino; Pilipino', code1='', code2='fil'),
    Language(name='Finnish', code1='fi', code2='fin'),
    Language(name='Finno-Ugrian languages', code1='', code2='fiu'),
    Language(name='Fon', code1='', code2='fon'),
    Language(name='French', code1='fr', code2='fra'),
    Language(name='French', code1='fr', code2='fre'),
    Language(name='French, Middle (ca.1400-1600)', code1='', code2='frm'),
    Language(name='French, Old (842-ca.1400)', code1='', code2='fro'),
    Language(name='Northern Frisian', code1='', code2='frr'),
    Language(name='Eastern Frisian', code1='', code2='frs'),
    Language(name='Western Frisian', code1='fy', code2='fry'),
    Language(name='Fulah', code1='ff', code2='ful'),
    Language(name='Friulian', code1='', code2='fur'),
    Language(name='Ga', code1='', code2='gaa'),
    Language(name='Gayo', code1='', code2='gay'),
    Language(name='Gbaya', code1='', code2='gba'),
    Language(name='Germanic languages', code1='', code2='gem'),
    Language(name='Georgian', code1='ka', code2='geo'),
    Language(name='German', code1='de', code2='ger'),
    Language(name='Geez', code1='', code2='gez'),
    Language(name='Gilbertese', code1='', code2='gil'),
    Language(name='Gaelic; Scottish Gaelic', code1='gd', code2='gla'),
    Language(name='Irish', code1='ga', code2='gle'),
    Language(name='Galician', code1='gl', code2='glg'),
    Language(name='Manx', code1='gv', code2='glv'),
    Language(name='German, Middle High (ca.1050-1500)', code1='', code2='gmh'),
    Language(name='German, Old High (ca.750-1050)', code1='', code2='goh'),
    Language(name='Gondi', code1='', code2='gon'),
    Language(name='Gorontalo', code1='', code2='gor'),
    Language(name='Gothic', code1='', code2='got'),
    Language(name='Grebo', code1='', code2='grb'),
    Language(name='Greek, Ancient (to 1453)', code1='', code2='grc'),
    Language(name='Greek, Modern (1453-)', code1='el', code2='gre'),
    Language(name='Guarani', code1='gn', code2='grn'),
    Language(name='Swiss German; Alemannic; Alsatian', code1='', code2='gsw'),
    Language(name='goudjrati', code1='Gujarati', code2='gujgu'),
    Language(name="Gwich'in", code1='', code2='gwi'),
    Language(name='Haida', code1='', code2='hai'),
    Language(name='Haitian; Haitian Creole', code1='ht', code2='hat'),
    Language(name='Hausa', code1='ha', code2='hau'),
    Language(name='Hawaiian', code1='', code2='haw'),
    Language(name='Hebrew', code1='he', code2='heb'),
    Language(name='Herero', code1='hz', code2='her'),
    Language(name='Hiligaynon', code1='', code2='hil'),
    Language(name='Himachali languages; Western Pahari languages', code1='', code2='him'),
    Language(name='Hindi', code1='hi', code2='hin'),
    Language(name='Hittite', code1='', code2='hit'),
    Language(name='Hmong; Mong', code1='', code2='hmn'),
    Language(name='Hiri Motu', code1='ho', code2='hmo'),
    Language(name='Croatian', code1='hr', code2='hrv'),
    Language(name='Upper Sorbian', code1='', code2='hsb'),
    Language(name='Hungarian', code1='hu', code2='hun'),
    Language(name='Hupa', code1='', code2='hup'),
    Language(name='Armenian', code1='hy', code2='hye'),
    Language(name='Iban', code1='', code2='iba'),
    Language(name='Igbo', code1='ig', code2='ibo'),
    Language(name='Icelandic', code1='is', code2='ice'),
    Language(name='Ido', code1='io', code2='ido'),
    Language(name='Sichuan Yi; Nuosu', code1='ii', code2='iii'),
    Language(name='Ijo languages', code1='', code2='ijo'),
    Language(name='Inuktitut', code1='iu', code2='iku'),
    Language(name='Interlingue; Occidental', code1='ie', code2='ile'),
    Language(name='Iloko', code1='', code2='ilo'),
    Language(name='Interlingua (International Auxiliary Language Association)', code1='ia', code2='ina'),
    Language(name='Indic languages', code1='', code2='inc'),
    Language(name='Indonesian', code1='id', code2='ind'),
    Language(name='Indo-European languages', code1='', code2='ine'),
    Language(name='Ingush', code1='', code2='inh'),
    Language(name='Inupiaq', code1='ik', code2='ipk'),
    Language(name='Iranian languages', code1='', code2='ira'),
    Language(name='Iroquoian languages', code1='', code2='iro'),
    Language(name='Icelandic', code1='is', code2='isl'),
    Language(name='Italian', code1='it', code2='ita'),
    Language(name='Javanese', code1='jv', code2='jav'),
    Language(name='Lojban', code1='', code2='jbo'),
    Language(name='Japanese', code1='ja', code2='jpn'),
    Language(name='Judeo-Persian', code1='', code2='jpr'),
    Language(name='Judeo-Arabic', code1='', code2='jrb'),
    Language(name='Kara-Kalpak', code1='', code2='kaa'),
    Language(name='Kabyle', code1='', code2='kab'),
    Language(name='Kachin; Jingpho', code1='', code2='kac'),
    Language(name='Kalaallisut; Greenlandic', code1='kl', code2='kal'),
    Language(name='Kamba', code1='', code2='kam'),
    Language(name='Kannada', code1='kn', code2='kan'),
    Language(name='Karen languages', code1='', code2='kar'),
    Language(name='Kashmiri', code1='ks', code2='kas'),
    Language(name='Georgian', code1='ka', code2='kat'),
    Language(name='Kanuri', code1='kr', code2='kau'),
    Language(name='Kawi', code1='', code2='kaw'),
    Language(name='Kazakh', code1='kk', code2='kaz'),
    Language(name='Kabardian', code1='', code2='kbd'),
    Language(name='Khasi', code1='', code2='kha'),
    Language(name='Khoisan languages', code1='', code2='khi'),
    Language(name='Central Khmer', code1='km', code2='khm'),
    Language(name='Khotanese; Sakan', code1='', code2='kho'),
    Language(name='Kikuyu; Gikuyu', code1='ki', code2='kik'),
    Language(name='Kinyarwanda', code1='rw', code2='kin'),
    Language(name='Kirghiz; Kyrgyz', code1='ky', code2='kir'),
    Language(name='Kimbundu', code1='', code2='kmb'),
    Language(name='Konkani', code1='', code2='kok'),
    Language(name='Komi', code1='kv', code2='kom'),
    Language(name='Kongo', code1='kg', code2='kon'),
    Language(name='Korean', code1='ko', code2='kor'),
    Language(name='Kosraean', code1='', code2='kos'),
    Language(name='Kpelle', code1='', code2='kpe'),
    Language(name='Karachay-Balkar', code1='', code2='krc'),
    Language(name='Karelian', code1='', code2='krl'),
    Language(name='Kru languages', code1='', code2='kro'),
    Language(name='Kurukh', code1='', code2='kru'),
    Language(name='Kuanyama; Kwanyama', code1='kj', code2='kua'),
    Language(name='Kumyk', code1='', code2='kum'),
    Language(name='Kurdish', code1='ku', code2='kur'),
    Language(name='Kutenai', code1='', code2='kut'),
    Language(name='Ladino', code1='', code2='lad'),
    Language(name='Lahnda', code1='', code2='lah'),
    Language(name='Lamba', code1='', code2='lam'),
    Language(name='Lao', code1='lo', code2='lao'),
    Language(name='Latin', code1='la', code2='lat'),
    Language(name='Latvian', code1='lv', code2='lav'),
    Language(name='Lezghian', code1='', code2='lez'),
    Language(name='Limburgan; Limburger; Limburgish', code1='li', code2='lim'),
    Language(name='Lingala', code1='ln', code2='lin'),
    Language(name='Lithuanian', code1='lt', code2='lit'),
    Language(name='Mongo', code1='', code2='lol'),
    Language(name='Lozi', code1='', code2='loz'),
    Language(name='Luxembourgish; Letzeburgesch', code1='lb', code2='ltz'),
    Language(name='Luba-Lulua', code1='', code2='lua'),
    Language(name='Luba-Katanga', code1='lu', code2='lub'),
    Language(name='Ganda', code1='lg', code2='lug'),
    Language(name='Luiseno', code1='', code2='lui'),
    Language(name='Lunda', code1='', code2='lun'),
    Language(name='luo (Kenya et Tanzanie)', code1=' Luo (Kenya and Tanzania)', code2='luo'),
    Language(name='Lushai', code1='', code2='lus'),
    Language(name='Macedonian', code1='mk', code2='mac'),
    Language(name='Madurese', code1='', code2='mad'),
    Language(name='Magahi', code1='', code2='mag'),
    Language(name='Marshallese', code1='mh', code2='mah'),
    Language(name='Maithili', code1='', code2='mai'),
    Language(name='Makasar', code1='', code2='mak'),
    Language(name='Malayalam', code1='ml', code2='mal'),
    Language(name='Mandingo', code1='', code2='man'),
    Language(name='Maori', code1='mi', code2='mao'),
    Language(name='Austronesian languages', code1='', code2='map'),
    Language(name='Marathi', code1='mr', code2='mar'),
    Language(name='Masai', code1='', code2='mas'),
    Language(name='Malay', code1='ms', code2='may'),
    Language(name='Moksha', code1='', code2='mdf'),
    Language(name='Mandar', code1='', code2='mdr'),
    Language(name='Mende', code1='', code2='men'),
    Language(name='Irish, Middle (900-1200)', code1='', code2='mga'),
    Language(name="Mi'kmaq; Micmac", code1='', code2='mic'),
    Language(name='Minangkabau', code1='', code2='min'),
    Language(name='Uncoded languages', code1='', code2='mis'),
    Language(name='Macedonian', code1='mk', code2='mkd'),
    Language(name='Mon-Khmer languages', code1='', code2='mkh'),
    Language(name='Malagasy', code1='mg', code2='mlg'),
    Language(name='Maltese', code1='mt', code2='mlt'),
    Language(name='Manchu', code1='', code2='mnc'),
    Language(name='Manipuri', code1='', code2='mni'),
    Language(name='Manobo languages', code1='', code2='mno'),
    Language(name='Mohawk', code1='', code2='moh'),
    Language(name='Mongolian', code1='mn', code2='mon'),
    Language(name='Mossi', code1='', code2='mos'),
    Language(name='Maori', code1='mi', code2='mri'),
    Language(name='Malay', code1='ms', code2='msa'),
    Language(name='Multiple languages', code1='', code2='mul'),
    Language(name='Munda languages', code1='', code2='mun'),
    Language(name='Creek', code1='', code2='mus'),
    Language(name='Mirandese', code1='', code2='mwl'),
    Language(name='Marwari', code1='', code2='mwr'),
    Language(name='Burmese', code1='my', code2='mya'),
    Language(name='Mayan languages', code1='', code2='myn'),
    Language(name='Erzya', code1='', code2='myv'),
    Language(name='Nahuatl languages', code1='', code2='nah'),
    Language(name='North American Indian languages', code1='', code2='nai'),
    Language(name='Neapolitan', code1='', code2='nap'),
    Language(name='Nauru', code1='na', code2='nau'),
    Language(name='Navajo; Navaho', code1='nv', code2='nav'),
    Language(name='Ndebele, South; South Ndebele', code1='nr', code2='nbl'),
    Language(name='Ndebele, North; North Ndebele', code1='nd', code2='nde'),
    Language(name='Ndonga', code1='ng', code2='ndo'),
    Language(name='Low German; Low Saxon; German, Low; Saxon, Low', code1='', code2='nds'),
    Language(name='Nepali', code1='ne', code2='nep'),
    Language(name='Nepal Bhasa; Newari', code1='', code2='new'),
    Language(name='Nias', code1='', code2='nia'),
    Language(name='Niger-Kordofanian languages', code1='', code2='nic'),
    Language(name='Niuean', code1='', code2='niu'),
    Language(name='Dutch; Flemish', code1='nl', code2='nld'),
    Language(name='Norwegian Nynorsk; Nynorsk, Norwegian', code1='nn', code2='nno'),
    Language(name='Bokmål, Norwegian; Norwegian Bokmål', code1='nb', code2='nob'),
    Language(name='Nogai', code1='', code2='nog'),
    Language(name='Norse, Old', code1='', code2='non'),
    Language(name='Norwegian', code1='no', code2='nor'),
    Language(name="N'Ko", code1='', code2='nqo'),
    Language(name='Pedi; Sepedi; Northern Sotho', code1='', code2='nso'),
    Language(name='Nubian languages', code1='', code2='nub'),
    Language(name='Classical Newari; Old Newari; Classical Nepal Bhasa', code1='', code2='nwc'),
    Language(name='Chichewa; Chewa; Nyanja', code1='ny', code2='nya'),
    Language(name='Nyamwezi', code1='', code2='nym'),
    Language(name='Nyankole', code1='', code2='nyn'),
    Language(name='Nyoro', code1='', code2='nyo'),
    Language(name='Nzima', code1='', code2='nzi'),
    Language(name='Occitan (post 1500)', code1='oc', code2='oci'),
    Language(name='Ojibwa', code1='oj', code2='oji'),
    Language(name='Oriya', code1='or', code2='ori'),
    Language(name='Oromo', code1='om', code2='orm'),
    Language(name='Osage', code1='', code2='osa'),
    Language(name='Ossetian; Ossetic', code1='os', code2='oss'),
    Language(name='Turkish, Ottoman (1500-1928)', code1='', code2='ota'),
    Language(name='Otomian languages', code1='', code2='oto'),
    Language(name='Papuan languages', code1='', code2='paa'),
    Language(name='Pangasinan', code1='', code2='pag'),
    Language(name='Pahlavi', code1=' ', code2='pal'),
    Language(name='Pampanga; Kapampangan', code1=' ', code2='pam'),
    Language(name='pendjabi', code1='paPanjabi; Punjabi', code2='pan'),
    Language(name='Papiamento', code1=' ', code2='pap'),
    Language(name='Palauan', code1=' ', code2='pau'),
    Language(name='Persian, Old (ca.600-400 B.C.)', code1=' ', code2='peo'),
    Language(name='Persian', code1='fa', code2='per'),
    Language(name='Philippine languages', code1=' ', code2='phi'),
    Language(name='Phoenician', code1=' ', code2='phn'),
    Language(name='Pali', code1='pi', code2='pli'),
    Language(name='Polish', code1='pl', code2='pol'),
    Language(name='Pohnpeian', code1=' ', code2='pon'),
    Language(name='Portuguese', code1='pt', code2='por'),
    Language(name='Prakrit languages', code1=' ', code2='pra'),
    Language(name='Provençal, Old (to 1500);Occitan, Old (to 1500)', code1=' ', code2='pro'),
    Language(name='Pushto; Pashto', code1='ps', code2='pus'),
    Language(name='Quechua', code1='qu', code2='que'),
    Language(name='Rajasthani', code1=' ', code2='raj'),
    Language(name='Rapanui', code1=' ', code2='rap'),
    Language(name='Rarotongan; Cook Islands Maori', code1=' ', code2='rar'),
    Language(name='Romance languages', code1=' ', code2='roa'),
    Language(name='Romansh', code1='rm', code2='roh'),
    Language(name='Romany', code1=' ', code2='rom'),
    Language(name='Romanian; Moldavian; Moldovan', code1='ro', code2='rum'),
    Language(name='Romanian; Moldavian; Moldovan', code1='ro', code2='ron'),
    Language(name='Rundi', code1='rn', code2='run'),
    Language(name='Aromanian; Arumanian; Macedo-Romanian', code1=' ', code2='rup'),
    Language(name='Russian', code1='ru', code2='rus'),
    Language(name='Sandawe', code1=' ', code2='sad'),
    Language(name='Sango', code1='sg', code2='sag'),
    Language(name='Yakut', code1=' ', code2='sah'),
    Language(name='South American Indian languages', code1=' ', code2='sai'),
    Language(name='Salishan languages', code1=' ', code2='sal'),
    Language(name='Samaritan Aramaic', code1=' ', code2='sam'),
    Language(name='Sanskrit', code1='sa', code2='san'),
    Language(name='Sasak', code1=' ', code2='sas'),
    Language(name='Santali', code1=' ', code2='sat'),
    Language(name='Sicilian', code1=' ', code2='scn'),
    Language(name='Scots', code1=' ', code2='sco'),
    Language(name='Selkup', code1=' ', code2='sel'),
    Language(name='Semitic languages', code1=' ', code2='sem'),
    Language(name='Irish, Old (to 900)', code1=' ', code2='sga'),
    Language(name='Sign Languages', code1=' ', code2='sgn'),
    Language(name='Shan', code1=' ', code2='shn'),
    Language(name='Sidamo', code1=' ', code2='sid'),
    Language(name='Sinhala; Sinhalese', code1='si', code2='sin'),
    Language(name='Siouan languages', code1=' ', code2='sio'),
    Language(name='Sino-Tibetan languages', code1=' ', code2='sit'),
    Language(name='Slavic languages', code1=' ', code2='sla'),
    Language(name='Slovak', code1='sk', code2='slo'),
    Language(name='Slovak', code1='sk', code2='slk'),
    Language(name='Slovenian', code1='sl', code2='slv'),
    Language(name='Southern Sami', code1=' ', code2='sma'),
    Language(name='Northern Sami', code1='se', code2='sme'),
    Language(name='Sami languages', code1=' ', code2='smi'),
    Language(name='Lule Sami', code1=' ', code2='smj'),
    Language(name='Inari Sami', code1=' ', code2='smn'),
    Language(name='Samoan', code1='sm', code2='smo'),
    Language(name='Skolt Sami', code1=' ', code2='sms'),
    Language(name='Shona', code1='sn', code2='sna'),
    Language(name='Sindhi', code1='sd', code2='snd'),
    Language(name='Soninke', code1=' ', code2='snk'),
    Language(name='Sogdian', code1=' ', code2='sog'),
    Language(name='Somali', code1='so', code2='som'),
    Language(name='Songhai languages', code1=' ', code2='son'),
    Language(name='Sotho, Southern', code1='st', code2='sot'),
    Language(name='Spanish', code1='es', code2='spa'),
    Language(name='Albanian', code1='sq', code2='sqi'),
    Language(name='Sardinian', code1='sc', code2='srd'),
    Language(name='Sranan Tongo', code1=' ', code2='srn'),
    Language(name='Serbian', code1='sr', code2='srp'),
    Language(name='Serer', code1=' ', code2='srr'),
    Language(name='Nilo-Saharan languages', code1=' ', code2='ssa'),
    Language(name='Swati', code1='ss', code2='ssw'),
    Language(name='Sukuma', code1=' ', code2='suk'),
    Language(name='Sundanese', code1='su', code2='sun'),
    Language(name='Susu', code1=' ', code2='sus'),
    Language(name='Sumerian', code1=' ', code2='sux'),
    Language(name='Swahili', code1='sw', code2='swa'),
    Language(name='Swedish', code1='sv', code2='swe'),
    Language(name='Classical Syriac', code1=' ', code2='syc'),
    Language(name='Syriac', code1=' ', code2='syr'),
    Language(name='Tahitian', code1='ty', code2='tah'),
    Language(name='Tai languages', code1=' ', code2='tai'),
    Language(name='Tamil', code1='ta', code2='tam'),
    Language(name='Tatar', code1='tt', code2='tat'),
    Language(name='Telugu', code1='te', code2='tel'),
    Language(name='Timne', code1=' ', code2='tem'),
    Language(name='Tereno', code1=' ', code2='ter'),
    Language(name='Tetum', code1=' ', code2='tet'),
    Language(name='Tajik', code1='tg', code2='tgk'),
    Language(name='Tagalog', code1='tl', code2='tgl'),
    Language(name='Thai', code1='th', code2='tha'),
    Language(name='Tibetan', code1='bo', code2='tib'),
    Language(name='Tigre', code1=' ', code2='tig'),
    Language(name='Tigrinya', code1='ti', code2='tir'),
    Language(name='Tiv', code1=' ', code2='tiv'),
    Language(name='Tokelau', code1=' ', code2='tkl'),
    Language(name='Klingon; tlhIngan-Hol', code1=' ', code2='tlh'),
    Language(name='Tlingit', code1=' ', code2='tli'),
    Language(name='Tamashek', code1=' ', code2='tmh'),
    Language(name='Tonga (Nyasa)', code1=' ', code2='tog'),
    Language(name='Tonga (Tonga Islands)', code1='to', code2='ton'),
    Language(name='Tok Pisin', code1=' ', code2='tpi'),
    Language(name='Tsimshian', code1=' ', code2='tsi'),
    Language(name='Tswana', code1='tn', code2='tsn'),
    Language(name='Tsonga', code1='ts', code2='tso'),
    Language(name='Turkmen', code1='tk', code2='tuk'),
    Language(name='Tumbuka', code1=' ', code2='tum'),
    Language(name='Tupi languages', code1=' ', code2='tup'),
    Language(name='Turkish', code1='tr', code2='tur'),
    Language(name='Altaic languages', code1=' ', code2='tut'),
    Language(name='Tuvalu', code1=' ', code2='tvl'),
    Language(name='Twi', code1='tw', code2='twi'),
    Language(name='Tuvinian', code1=' ', code2='tyv'),
    Language(name='Udmurt', code1=' ', code2='udm'),
    Language(name='Ugaritic', code1=' ', code2='uga'),
    Language(name='Uighur; Uyghur', code1='ug', code2='uig'),
    Language(name='Ukrainian', code1='uk', code2='ukr'),
    Language(name='Umbundu', code1=' ', code2='umb'),
    Language(name='Undetermined', code1=' ', code2='und'),
    Language(name='Urdu', code1='ur', code2='urd'),
    Language(name='Uzbek', code1='uz', code2='uzb'),
    Language(name='Vai', code1=' ', code2='vai'),
    Language(name='Venda', code1='ve', code2='ven'),
    Language(name='Vietnamese', code1='vi', code2='vie'),
    Language(name='Volapük', code1='vo', code2='vol'),
    Language(name='Votic', code1=' ', code2='vot'),
    Language(name='Wakashan languages', code1=' ', code2='wak'),
    Language(name='Wolaitta; Wolaytta', code1=' ', code2='wal'),
    Language(name='Waray', code1=' ', code2='war'),
    Language(name='Washo', code1=' ', code2='was'),
    Language(name='Welsh', code1='cy', code2='wel'),
    Language(name='Sorbian languages', code1=' ', code2='wen'),
    Language(name='Walloon', code1='wa', code2='wln'),
    Language(name='Wolof', code1='wo', code2='wol'),
    Language(name='Kalmyk; Oirat', code1=' ', code2='xal'),
    Language(name='Xhosa', code1='xh', code2='xho'),
    Language(name='Yao', code1=' ', code2='yao'),
    Language(name='Yapese', code1=' ', code2='yap'),
    Language(name='Yiddish', code1='yi', code2='yid'),
    Language(name='Yoruba', code1='yo', code2='yor'),
    Language(name='Yupik languages', code1=' ', code2='ypk'),
    Language(name='Zapotec', code1=' ', code2='zap'),
    Language(name='Blissymbols; Blissymbolics; Bliss', code1=' ', code2='zbl'),
    Language(name='Zenaga', code1=' ', code2='zen'),
    Language(name='Standard Moroccan Tamazight', code1=' ', code2='zgh'),
    Language(name='Zhuang; Chuang', code1='za', code2='zha'),
    Language(name='Chinese', code1='zh', code2='zho'),
    Language(name='Zande languages', code1=' ', code2='znd'),
    Language(name='Zulu', code1='zu', code2='zul'),
    Language(name='Zuni', code1=' ', code2='zun'),
    Language(name='Zaza; Dimili; Dimli; Kirdki; Kirmanjki; Zazaki', code1='', code2='zza'),
)

#: Mapping `code1` to language instance. Use ``LANGUAGES_MAP_CODE1.get('someCode1')``. ``None`` means unknown lang.
LANGUAGES_MAP_CODE1: t.Mapping[str, Language] = MappingProxyType({lang.code1: lang for lang in LANGUAGES})

#: Mapping `code2` to language instance. Use ``LANGUAGES_MAP_CODE2.get('someCode2')``. ``None`` means unknown lang.
LANGUAGES_MAP_CODE2: t.Mapping[str, Language] = MappingProxyType({lang.code2: lang for lang in LANGUAGES})

#: Default language code1 used by SDK.
DEFAULT_LANGUAGE_CODE1 = 'en'
#: Default :obj:`Language` instance used by SDK.
DEFAULT_LANGUAGE: Language = LANGUAGES_MAP_CODE1[DEFAULT_LANGUAGE_CODE1]



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

import typing_extensions as te
from pydantic import Field

from atproto_client.models import base


class Data(base.DataModelBase):
    """Input data model for :obj:`com.atproto.admin.sendEmail`."""

    content: str  #: Content.
    recipient_did: str = Field(alias='recipientDid')  #: Recipient did.
    sender_did: str = Field(alias='senderDid')  #: Sender did.
    comment: t.Optional[
        str
    ] = None  #: Additional comment by the sender that won't be used in the email itself but helpful to provide more context for moderators/reviewers.
    subject: t.Optional[str] = None  #: Subject.


class DataDict(te.TypedDict):
    content: str  #: Content.
    recipient_did: str  #: Recipient did.
    sender_did: str  #: Sender did.
    comment: te.NotRequired[
        t.Optional[str]
    ]  #: Additional comment by the sender that won't be used in the email itself but helpful to provide more context for moderators/reviewers.
    subject: te.NotRequired[t.Optional[str]]  #: Subject.


class Response(base.ResponseModelBase):
    """Output data model for :obj:`com.atproto.admin.sendEmail`."""

    sent: bool  #: Sent.



import asyncio
import typing as t
from datetime import timedelta

from atproto_server.auth.jwt import get_jwt_payload

from atproto_client.client.session import (
    AsyncSessionChangeCallback,
    Session,
    SessionChangeCallback,
    SessionEvent,
    SessionResponse,
)

if t.TYPE_CHECKING:
    from atproto_server.auth.jwt import JwtPayload


class SessionDispatchMixin:
    def __init__(self, *args, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)

        self._on_session_change_callbacks: t.List[SessionChangeCallback] = []
        self._on_session_change_async_callbacks: t.List[AsyncSessionChangeCallback] = []

    def on_session_change(self, callback: SessionChangeCallback) -> None:
        """Register a callback for session change event.

        Args:
            callback: A callback to be called when the session changes.
                The callback must accept two arguments: event and session.

        Example:
            >>> from atproto import Client, SessionEvent, Session
            >>>
            >>> client = Client()
            >>>
            >>> def on_session_change(event: SessionEvent, session: Session):
            >>>     print(event, session)
            >>>
            >>> client.on_session_change(on_session_change)

        Returns:
            :obj:`None`
        """
        self._on_session_change_callbacks.append(callback)

    def _call_on_session_change_callbacks(self, event: SessionEvent, session: Session) -> None:
        for on_session_change_callback in self._on_session_change_callbacks:
            on_session_change_callback(event, session)


class AsyncSessionDispatchMixin:
    def __init__(self, *args, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)

        self._on_session_change_async_callbacks: t.List[AsyncSessionChangeCallback] = []

    def on_session_change(self, callback: AsyncSessionChangeCallback) -> None:
        """Register a callback for session change event.

        Args:
            callback: A callback to be called when the session changes.
                The callback must accept two arguments: event and session.

        Note:
            Possible events: `SessionEvent.IMPORT`, `SessionEvent.CREATE`, `SessionEvent.REFRESH`.

        Tip:
            You should save the session string to persistent storage
            on `SessionEvent.CREATE` and `SessionEvent.REFRESN` event.

        Example:
            >>> from atproto import AsyncClient, SessionEvent, Session
            >>>
            >>> client = AsyncClient()
            >>>
            >>> async def on_session_change(event: SessionEvent, session: Session):
            >>>     print(event, session)
            >>>
            >>> client.on_session_change(on_session_change)

        Returns:
            :obj:`None`
        """
        self._on_session_change_async_callbacks.append(callback)

    async def _call_on_session_change_callbacks(self, event: SessionEvent, session: Session) -> None:
        coroutines = []
        for on_session_change_async_callback in self._on_session_change_async_callbacks:
            coroutines.append(on_session_change_async_callback(event, session))

        await asyncio.gather(*coroutines)


class SessionMethodsMixin:
    def __init__(self, *args, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)

        self._access_jwt: t.Optional[str] = None
        self._access_jwt_payload: t.Optional['JwtPayload'] = None

        self._refresh_jwt: t.Optional[str] = None
        self._refresh_jwt_payload: t.Optional['JwtPayload'] = None

        self._session: t.Optional[Session] = None

    def _should_refresh_session(self) -> bool:
        expired_at = self.get_time_from_timestamp(self._access_jwt_payload.exp)
        expired_at = expired_at - timedelta(minutes=15)  # let's update the token a bit earlier than required

        return self.get_current_time() > expired_at

    def _set_session_common(self, session: SessionResponse) -> None:
        self._access_jwt = session.access_jwt
        self._access_jwt_payload = get_jwt_payload(session.access_jwt)

        self._refresh_jwt = session.refresh_jwt
        self._refresh_jwt_payload = get_jwt_payload(session.refresh_jwt)

        self._session = Session(
            access_jwt=session.access_jwt,
            refresh_jwt=session.refresh_jwt,
            did=session.did,
            handle=session.handle,
        )

        self._set_auth_headers(session.access_jwt)

    @staticmethod
    def _get_auth_headers(token: str) -> t.Dict[str, str]:
        return {'Authorization': f'Bearer {token}'}

    def _set_auth_headers(self, token: str) -> None:
        self.request.set_additional_headers(self._get_auth_headers(token))

    def export_session_string(self) -> str:
        """Export session string.

        Note:
            This method is useful for storing the session and reusing it later.

        Warning:
            You should use it if you create the client instance often.
            Because of server rate limits for `createSession`.
            Rate limited by handle.
            30/5 min, 300/day.

        Attention:
            You must export session at the end of the Client`s life cycle!
            Alternatively, you can subscribe to the session change event.
            Use :py:attr:`~on_session_change` to register handler.

        Example:
            >>> from atproto import Client
            >>> # the first time login with login and password
            >>> client = Client()
            >>> client.login('login', 'password')
            >>> session_string = client.export_session_string()
            >>> # store session_string somewhere.
            >>> # for example, in env and next time use it for login
            >>> client2 = Client()
            >>> client2.login(session_string=session_string)

        Returns:
            :obj:`str`: Session string.
        """
        return self._session.export()



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

import typing_extensions as te
from pydantic import Field

if t.TYPE_CHECKING:
    from atproto_client import models
from atproto_client.models import base


class Params(base.ParamsModelBase):
    """Parameters model for :obj:`com.atproto.label.queryLabels`."""

    uri_patterns: t.List[str] = Field(
        alias='uriPatterns'
    )  #: List of AT URI patterns to match (boolean 'OR'). Each may be a prefix (ending with '*'; will match inclusive of the string leading to '*'), or a full URI.
    cursor: t.Optional[str] = None  #: Cursor.
    limit: t.Optional[int] = Field(default=50, ge=1, le=250)  #: Limit.
    sources: t.Optional[t.List[str]] = None  #: Optional list of label sources (DIDs) to filter on.


class ParamsDict(te.TypedDict):
    uri_patterns: t.List[
        str
    ]  #: List of AT URI patterns to match (boolean 'OR'). Each may be a prefix (ending with '*'; will match inclusive of the string leading to '*'), or a full URI.
    cursor: te.NotRequired[t.Optional[str]]  #: Cursor.
    limit: te.NotRequired[t.Optional[int]]  #: Limit.
    sources: te.NotRequired[t.Optional[t.List[str]]]  #: Optional list of label sources (DIDs) to filter on.


class Response(base.ResponseModelBase):
    """Output data model for :obj:`com.atproto.label.queryLabels`."""

    labels: t.List['models.ComAtprotoLabelDefs.Label']  #: Labels.
    cursor: t.Optional[str] = None  #: Cursor.



from typing import List

import libipld

from atproto_core.exceptions import DAGCBORDecodingError


def decode_dag(data: bytes) -> dict:
    """Decode and returns a single data item from the given data, with the DAG-CBOR codec.

    Args:
        data: The stream of bytes or bytes with DAG-CBOR data.

    Returns:
        :obj:`dict`: Decoded DAG-CBOR.
    """
    try:
        return libipld.decode_dag_cbor(data)
    except Exception as e:  # noqa: BLE001
        raise DAGCBORDecodingError from e


def decode_dag_multi(data: bytes) -> List[dict]:
    """Decode and returns many data items from the given data, with the DAG-CBOR codec.

    Args:
        data: The stream of bytes or bytes with DAG-CBOR data.

    Returns:
        :obj:`list` of :obj:`dict`: Decoded DAG-CBOR.
    """
    return libipld.decode_dag_cbor_multi(data)



import typing as t

from atproto_identity.did.resolver import AsyncDidResolver, DidResolver
from atproto_identity.handle.resolver import AsyncHandleResolver, HandleResolver

if t.TYPE_CHECKING:
    from atproto_identity.cache.base_cache import AsyncDidBaseCache, DidBaseCache


class IdResolver:
    """Identity Resolver.

    This resolver is used to resolve identities.
    DID and Handle identifies are supported.

    Note:
        Default PLC directory URL is https://plc.directory.
        Default request timeout is 3 seconds.

    Args:
        plc_url: PLC directory URL.
        timeout: Request timeout.
        cache: DID cache.
    """

    def __init__(
        self,
        plc_url: t.Optional[str] = None,
        timeout: t.Optional[float] = None,
        cache: t.Optional['DidBaseCache'] = None,
        backup_nameservers: t.Optional[t.List[str]] = None,
    ) -> None:
        self._handle = HandleResolver(timeout, backup_nameservers)
        self._did = DidResolver(plc_url, timeout, cache)

    @property
    def handle(self) -> HandleResolver:
        """Handle Resolver.

        This resolver is used to resolve handles.
        """
        return self._handle

    @property
    def did(self) -> DidResolver:
        """DID Resolver.

        This resolver is used to resolve DIDs.
        PLC and Web DID methods are supported.
        """
        return self._did


class AsyncIdResolver:
    """Asynchronous Identity Resolver.

    This resolver is used to resolve identities.
    DID and Handle identifies are supported.

    Note:
        Default PLC directory URL is https://plc.directory.
        Default request timeout is 3 seconds.

    Args:
        plc_url: PLC directory URL.
        timeout: Request timeout.
        cache: DID cache.
    """

    def __init__(
        self,
        plc_url: t.Optional[str] = None,
        timeout: t.Optional[float] = None,
        cache: t.Optional['AsyncDidBaseCache'] = None,
        backup_nameservers: t.Optional[t.List[str]] = None,
    ) -> None:
        self._handle = AsyncHandleResolver(timeout, backup_nameservers)
        self._did = AsyncDidResolver(plc_url, timeout, cache)

    @property
    def handle(self) -> AsyncHandleResolver:
        """Handle Resolver.

        This resolver is used to resolve handles.
        """
        return self._handle

    @property
    def did(self) -> AsyncDidResolver:
        """DID Resolver.

        This resolver is used to resolve DIDs.
        PLC and Web DID methods are supported.
        """
        return self._did



import typing as t
from io import BytesIO

from atproto_client import models

_FEATURES_TYPE = t.List[
    t.Union[
        'models.AppBskyRichtextFacet.Mention',
        'models.AppBskyRichtextFacet.Link',
        'models.AppBskyRichtextFacet.Tag',
    ]
]


class TextBuilder:
    """Helper that helps construct rich text.

    There are three facets: link, mention, tag.

    Each facet could be linked to text segment.

    This helper provides chained API.

    Example:
        Without chaining:

        >>> from atproto import client_utils
        >>> text_builder = client_utils.TextBuilder()
        >>> text_builder.tag('This is a rich message. ', 'atproto')
        >>> text_builder.text('I can mention ')
        >>> text_builder.mention('account', 'did:plc:kvwvcn5iqfooopmyzvb4qzba')
        >>> text_builder.text(' and add clickable ')
        >>> text_builder.link('link', 'https://atproto.blue/')

        With chaining:

        >>> from atproto import client_utils
        >>> text_builder = client_utils.TextBuilder().text('Test msg using ').link('Python SDK', 'https://atproto.blue/')

        Later you can use this builder in the Client:

        >>> from atproto import Client, client_utils
        >>> client = Client()
        >>> # You can pass instance of TextBuilder instead of str to the "text" argument.
        >>> client.send_post(client_utils.TextBuilder().link('Python SDK', 'https://atproto.blue/'))
        >>> # Same with send_image method

    Note:
        This helper doesn't support overlapped features (features at the same byte range).
    """

    def __init__(self) -> None:
        self._buffer = BytesIO()
        self._facets: t.List[models.AppBskyRichtextFacet.Main] = []

    @property
    def __tell(self) -> int:
        return self._buffer.tell()

    _start_index = __tell
    _end_index = __tell

    def _write_text(self, text: str) -> t.Tuple[int, int]:
        start_index = self._start_index
        self._write_bytes(text.encode('UTF-8'))
        end_index = self._end_index
        return start_index, end_index

    def _write_bytes(self, data: bytes) -> int:
        return self._buffer.write(data)

    def _add_facet(self, facet: models.AppBskyRichtextFacet.Main) -> None:
        self._facets.append(facet)

    def _create_byte_slice(self, text: str) -> models.AppBskyRichtextFacet.ByteSlice:
        start, end = self._write_text(text)
        return models.AppBskyRichtextFacet.ByteSlice(byte_start=start, byte_end=end)

    def _create_facet(self, text: str, features: _FEATURES_TYPE) -> models.AppBskyRichtextFacet.Main:
        return models.AppBskyRichtextFacet.Main(
            features=features,
            index=self._create_byte_slice(text),
        )

    def build_facets(self) -> t.List[models.AppBskyRichtextFacet.Main]:
        """Build facets from the current state of the builder."""
        return self._facets.copy()

    def build_text(self) -> str:
        """Build text from the current state of the builder."""
        return self._buffer.getvalue().decode()

    def text(self, text: str) -> 'TextBuilder':
        """Add text to the builder.

        Args:
            text: Text.
        """
        self._write_text(text)
        return self

    def link(self, text: str, url: str) -> 'TextBuilder':
        """Add link to the builder.

        Args:
            text: Text of the link.
            url: Valid URL. For example, https://atproto.blue/
        """
        self._add_facet(self._create_facet(text, [models.AppBskyRichtextFacet.Link(uri=url)]))
        return self

    def mention(self, text: str, did: str) -> 'TextBuilder':
        """Add mention to the builder.

        Args:
            text: Text of the mention.
            did: Valid DID. For example, did:plc:kvwvcn5iqfooopmyzvb4qzba
        """
        self._add_facet(self._create_facet(text, [models.AppBskyRichtextFacet.Mention(did=did)]))
        return self

    def tag(self, text: str, tag: str) -> 'TextBuilder':
        """Add tag to the builder.

        Args:
            text: Text of the tag.
            tag: Hashtag.
        """
        self._add_facet(self._create_facet(text, [models.AppBskyRichtextFacet.Tag(tag=tag)]))
        return self



import typing as t
from datetime import datetime, timezone

from atproto_core.did_doc import DidDocument

from atproto_identity.cache.base_cache import AsyncDidBaseCache, DidBaseCache
from atproto_identity.cache.models import CachedDid, CachedDidResult

if t.TYPE_CHECKING:
    from atproto_identity.cache.base_cache import AsyncGetDocCallback, GetDocCallback


def _datetime_now() -> datetime:
    return datetime.now(timezone.utc)


class DidInMemoryCache(DidBaseCache):
    def __init__(self, *args, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)

        self._cache: t.Dict[str, CachedDid] = {}

    def set(self, did: str, document: DidDocument) -> None:
        self._cache[did] = CachedDid(document, _datetime_now())

    def refresh(self, did: str, get_doc_callback: 'GetDocCallback') -> None:
        doc = get_doc_callback()
        if doc:
            self.set(did, doc)

    def delete(self, did: str) -> None:
        del self._cache[did]

    def clear(self) -> None:
        self._cache.clear()

    def get(self, did: str) -> t.Optional[CachedDidResult]:
        val = self._cache.get(did)
        if not val:
            return None

        now = _datetime_now().timestamp()
        expired = now > val.updated_at.timestamp() + self.max_ttl
        stale = now > val.updated_at.timestamp() + self.stale_ttl

        return CachedDidResult(did, val.document, val.updated_at, stale, expired)


class AsyncDidInMemoryCache(AsyncDidBaseCache):
    def __init__(self, *args, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)

        self._in_memory_cache = DidInMemoryCache(*args, **kwargs)

    async def set(self, did: str, document: DidDocument) -> None:
        self._in_memory_cache.set(did, document)

    async def refresh(self, did: str, get_doc_callback: 'AsyncGetDocCallback') -> None:
        doc = await get_doc_callback()
        if doc:
            await self.set(did, doc)

    async def delete(self, did: str) -> None:
        self._in_memory_cache.delete(did)

    async def clear(self) -> None:
        self._in_memory_cache.clear()

    async def get(self, did: str) -> t.Optional[CachedDidResult]:
        return self._in_memory_cache.get(did)



from .firehose import (
    AsyncFirehoseSubscribeLabelsClient,
    AsyncFirehoseSubscribeReposClient,
    FirehoseSubscribeLabelsClient,
    FirehoseSubscribeReposClient,
    parse_subscribe_labels_message,
    parse_subscribe_repos_message,
)

__all__ = [
    'AsyncFirehoseSubscribeLabelsClient',
    'AsyncFirehoseSubscribeReposClient',
    'FirehoseSubscribeLabelsClient',
    'FirehoseSubscribeReposClient',
    'parse_subscribe_labels_message',
    'parse_subscribe_repos_message',
]



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

import typing_extensions as te
from pydantic import Field

if t.TYPE_CHECKING:
    from atproto_client import models
from atproto_client.models import base


class ReplyRef(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.post`."""

    parent: 'models.ComAtprotoRepoStrongRef.Main'  #: Parent.
    root: 'models.ComAtprotoRepoStrongRef.Main'  #: Root.

    py_type: te.Literal['app.bsky.feed.post#replyRef'] = Field(
        default='app.bsky.feed.post#replyRef', alias='$type', frozen=True
    )


class Entity(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.post`. Deprecated: use facets instead."""

    index: 'models.AppBskyFeedPost.TextSlice'  #: Index.
    type: str  #: Expected values are 'mention' and 'link'.
    value: str  #: Value.

    py_type: te.Literal['app.bsky.feed.post#entity'] = Field(
        default='app.bsky.feed.post#entity', alias='$type', frozen=True
    )


class TextSlice(base.ModelBase):
    """Definition model for :obj:`app.bsky.feed.post`. Deprecated. Use app.bsky.richtext instead -- A text segment. Start is inclusive, end is exclusive. Indices are for utf16-encoded strings."""

    end: int = Field(ge=0)  #: End.
    start: int = Field(ge=0)  #: Start.

    py_type: te.Literal['app.bsky.feed.post#textSlice'] = Field(
        default='app.bsky.feed.post#textSlice', alias='$type', frozen=True
    )


class Record(base.RecordModelBase):
    """Record model for :obj:`app.bsky.feed.post`."""

    created_at: str = Field(alias='createdAt')  #: Created at.
    text: str = Field(max_length=3000)  #: Text.
    embed: t.Optional[
        te.Annotated[
            t.Union[
                'models.AppBskyEmbedImages.Main',
                'models.AppBskyEmbedExternal.Main',
                'models.AppBskyEmbedRecord.Main',
                'models.AppBskyEmbedRecordWithMedia.Main',
            ],
            Field(default=None, discriminator='py_type'),
        ]
    ] = None  #: Embed.
    entities: t.Optional[
        t.List['models.AppBskyFeedPost.Entity']
    ] = None  #: Deprecated: replaced by app.bsky.richtext.facet.
    facets: t.Optional[t.List['models.AppBskyRichtextFacet.Main']] = None  #: Facets.
    labels: t.Optional[
        te.Annotated[t.Union['models.ComAtprotoLabelDefs.SelfLabels'], Field(default=None, discriminator='py_type')]
    ] = None  #: Labels.
    langs: t.Optional[t.List[str]] = Field(default=None, max_length=3)  #: Langs.
    reply: t.Optional['models.AppBskyFeedPost.ReplyRef'] = None  #: Reply.
    tags: t.Optional[t.List[str]] = Field(
        default=None, max_length=8
    )  #: Additional non-inline tags describing this post.

    py_type: te.Literal['app.bsky.feed.post'] = Field(default='app.bsky.feed.post', alias='$type', frozen=True)


class Main(Record):
    def __init_subclass__(cls, **data: t.Any) -> None:
        import warnings

        warnings.warn('Main class is deprecated. Use Record class instead.', DeprecationWarning, stacklevel=2)
        super().__init_subclass__(**data)

    def __init__(self, **data: t.Any) -> None:
        import warnings

        warnings.warn('Main class is deprecated. Use Record class instead.', DeprecationWarning, stacklevel=2)
        super().__init__(**data)


class GetRecordResponse(base.SugarResponseModelBase):
    """Get record response for :obj:`models.AppBskyFeedPost.Record`."""

    uri: str  #: The URI of the record.
    value: 'models.AppBskyFeedPost.Record'  #: The record.
    cid: t.Optional[str] = None  #: The CID of the record.


class ListRecordsResponse(base.SugarResponseModelBase):
    """List records response for :obj:`models.AppBskyFeedPost.Record`."""

    records: t.Dict[str, 'models.AppBskyFeedPost.Record']  #: Map of URIs to records.
    cursor: t.Optional[str] = None  #: Next page cursor.


class CreateRecordResponse(base.SugarResponseModelBase):
    """Create record response for :obj:`models.AppBskyFeedPost.Record`."""

    uri: str  #: The URI of the record.
    cid: str  #: The CID of the record.



import typing as t

from atproto_identity.did.resolvers.base_resolver import AsyncBaseResolver, BaseResolver
from atproto_identity.did.resolvers.plc_resolver import AsyncDidPlcResolver, DidPlcResolver
from atproto_identity.did.resolvers.web_resolver import AsyncDidWebResolver, DidWebResolver
from atproto_identity.exceptions import PoorlyFormattedDidError, UnsupportedDidMethodError

if t.TYPE_CHECKING:
    from atproto_identity.cache.base_cache import AsyncDidBaseCache, DidBaseCache


_BASE_PLC_URL = 'https://plc.directory'
_DEFAULT_TIMEOUT = 3.0
_DID_PREFIX = 'did'


class _DidResolverBase:
    def __init__(self) -> None:
        self._methods: t.Dict[str, t.Union[BaseResolver, AsyncBaseResolver]] = {}

    def _get_resolver_method(self, did: str) -> t.Union[BaseResolver, AsyncBaseResolver]:
        parts = did.split(':')
        if not parts:
            raise PoorlyFormattedDidError(f'Invalid DID {did}')

        if parts[0] != _DID_PREFIX:
            raise PoorlyFormattedDidError(f'Invalid DID {did}')

        if len(parts) < 2:
            raise PoorlyFormattedDidError(f'Invalid DID {did}')

        method = parts[1]
        if method not in self._methods:
            raise UnsupportedDidMethodError(f'Invalid DID {did}')

        return self._methods[method]


class DidResolver(_DidResolverBase, BaseResolver):
    """DID Resolver.

    Supported DID methods: PLC, Web.

    Args:
        plc_url: PLC directory URL.
        timeout: Request timeout.
        cache: DID cache.
    """

    def __init__(
        self,
        plc_url: t.Optional[str] = None,
        timeout: t.Optional[float] = None,
        cache: t.Optional['DidBaseCache'] = None,
    ) -> None:
        super().__init__()
        BaseResolver.__init__(self, cache)

        if plc_url is None:
            plc_url = _BASE_PLC_URL

        if timeout is None:
            timeout = _DEFAULT_TIMEOUT

        self._methods = {'plc': DidPlcResolver(plc_url, timeout, cache), 'web': DidWebResolver(timeout)}

    def resolve_without_validation(self, did: str) -> t.Optional[dict]:
        """Resolve DID without validation.

        Args:
            did: DID.

        Returns:
            :obj:`dict`: DID document or ``None`` if DID not found.
        """
        return self._get_resolver_method(did).resolve_without_validation(did)


class AsyncDidResolver(_DidResolverBase, AsyncBaseResolver):
    """Asynchronous DID Resolver.

    Supported DID methods: PLC, Web.

    Args:
        plc_url: PLC directory URL.
        timeout: Request timeout.
        cache: DID cache.
    """

    def __init__(
        self,
        plc_url: t.Optional[str] = None,
        timeout: t.Optional[float] = None,
        cache: t.Optional['AsyncDidBaseCache'] = None,
    ) -> None:
        super().__init__()
        AsyncBaseResolver.__init__(self, cache)

        if plc_url is None:
            plc_url = _BASE_PLC_URL

        if timeout is None:
            timeout = _DEFAULT_TIMEOUT

        self._methods = {'plc': AsyncDidPlcResolver(plc_url, timeout, cache), 'web': AsyncDidWebResolver(timeout)}

    async def resolve_without_validation(self, did: str) -> t.Optional[dict]:
        """Resolve DID without validation.

        Args:
            did: DID.

        Returns:
            :obj:`dict`: DID document or ``None`` if DID not found.
        """
        return await self._get_resolver_method(did).resolve_without_validation(did)



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t

import typing_extensions as te
from pydantic import Field

if t.TYPE_CHECKING:
    from atproto_client import models
    from atproto_client.models.blob_ref import BlobRef
from atproto_client.models import base


class Main(base.ModelBase):
    """Definition model for :obj:`app.bsky.embed.images`."""

    images: t.List['models.AppBskyEmbedImages.Image'] = Field(max_length=4)  #: Images.

    py_type: te.Literal['app.bsky.embed.images'] = Field(default='app.bsky.embed.images', alias='$type', frozen=True)


class Image(base.ModelBase):
    """Definition model for :obj:`app.bsky.embed.images`."""

    alt: str  #: Alt.
    image: 'BlobRef'  #: Image.
    aspect_ratio: t.Optional['models.AppBskyEmbedImages.AspectRatio'] = Field(
        default=None, alias='aspectRatio'
    )  #: Aspect ratio.

    py_type: te.Literal['app.bsky.embed.images#image'] = Field(
        default='app.bsky.embed.images#image', alias='$type', frozen=True
    )


class AspectRatio(base.ModelBase):
    """Definition model for :obj:`app.bsky.embed.images`. width:height represents an aspect ratio. It may be approximate, and may not correspond to absolute dimensions in any given unit."""

    height: int = Field(ge=1)  #: Height.
    width: int = Field(ge=1)  #: Width.

    py_type: te.Literal['app.bsky.embed.images#aspectRatio'] = Field(
        default='app.bsky.embed.images#aspectRatio', alias='$type', frozen=True
    )


class View(base.ModelBase):
    """Definition model for :obj:`app.bsky.embed.images`."""

    images: t.List['models.AppBskyEmbedImages.ViewImage'] = Field(max_length=4)  #: Images.

    py_type: te.Literal['app.bsky.embed.images#view'] = Field(
        default='app.bsky.embed.images#view', alias='$type', frozen=True
    )


class ViewImage(base.ModelBase):
    """Definition model for :obj:`app.bsky.embed.images`."""

    alt: str  #: Alt.
    fullsize: str  #: Fullsize.
    thumb: str  #: Thumb.
    aspect_ratio: t.Optional['models.AppBskyEmbedImages.AspectRatio'] = Field(
        default=None, alias='aspectRatio'
    )  #: Aspect ratio.

    py_type: te.Literal['app.bsky.embed.images#viewImage'] = Field(
        default='app.bsky.embed.images#viewImage', alias='$type', frozen=True
    )



import os
import typing as t
from enum import Enum
from functools import lru_cache
from pathlib import Path

from atproto_core.nsid import NSID
from atproto_lexicon import models

from atproto_codegen.consts import (
    DISCLAIMER,
    INPUT_DICT,
    INPUT_MODEL,
    OUTPUT_MODEL,
    PARAMS_DICT,
    PARAMS_MODEL,
)
from atproto_codegen.models import builder
from atproto_codegen.record_templates import (
    RECORD_CREATE_RESPONSE_MODEL_TEMPLATE,
    RECORD_GET_RESPONSE_MODEL_TEMPLATE,
    RECORD_LIST_RESPONSE_MODEL_TEMPLATE,
)
from atproto_codegen.utils import (
    _resolve_nsid_ref,
    append_code,
    convert_camel_case_to_snake_case,
    format_code,
    gen_description_by_camel_case_name,
    get_def_model_name,
    get_file_path_parts,
    get_import_path,
    get_record_model_name,
    join_code,
    write_code,
)
from atproto_codegen.utils import get_code_intent as _

_MODELS_OUTPUT_DIR = Path(__file__).parent.parent.parent.joinpath('atproto_client', 'models')


class ModelType(Enum):
    PARAMS = 'Parameters'
    DATA = 'Input data'
    RESPONSE = 'Output data'
    DEF = 'Definition'
    RECORD = 'Record'


class TypedDictType(Enum):
    PARAMS = 'Parameters'
    DATA = 'Input data'


def save_code(nsid: NSID, code: str) -> None:
    path_to_file = _MODELS_OUTPUT_DIR.joinpath(*get_file_path_parts(nsid))
    write_code(_MODELS_OUTPUT_DIR.joinpath(path_to_file), code)


def save_code_part(nsid: NSID, code: str) -> None:
    path_to_file = _MODELS_OUTPUT_DIR.joinpath(*get_file_path_parts(nsid))
    append_code(_MODELS_OUTPUT_DIR.joinpath(path_to_file), code)


def _get_model_imports() -> str:
    # we are using ruff with F401 autofix to delete unused imports
    lines = [
        'import typing as t',
        '',
        'import typing_extensions as te',
        'from pydantic import Field',
        '',
        'if t.TYPE_CHECKING:',
        f'{_(1)}from atproto_client import models',
        f'{_(1)}from atproto_client.models.unknown_type import UnknownType',
        f'{_(1)}from atproto_client.models.unknown_type import UnknownInputType',
        f'{_(1)}from atproto_client.models.blob_ref import BlobRef',
        f'{_(1)}from atproto_core.cid import CIDType',
        'from atproto_client.models import base',
        '',
        '',
    ]

    return join_code(lines)


_NSID_WITH_IMPORTS = set()


def _save_code_import_if_not_exist(nsid: NSID) -> None:
    if nsid not in _NSID_WITH_IMPORTS:
        lines = [DISCLAIMER, _get_model_imports()]
        save_code(nsid, join_code(lines))
        _NSID_WITH_IMPORTS.add(nsid)


def _get_model_class_def(name: str, model_type: ModelType) -> str:
    lines: t.List[str] = []

    if model_type is ModelType.PARAMS:
        lines.append(f'class {PARAMS_MODEL}(base.ParamsModelBase):')
    elif model_type is ModelType.DATA:
        lines.append(f'class {INPUT_MODEL}(base.DataModelBase):')
    elif model_type is ModelType.RESPONSE:
        lines.append(f'class {OUTPUT_MODEL}(base.ResponseModelBase):')
    elif model_type is ModelType.RECORD:
        lines.append(f'class {get_record_model_name(name)}(base.RecordModelBase):')
    elif model_type is ModelType.DEF:
        lines.append(f'class {get_def_model_name(name)}(base.ModelBase):')

    lines.append('')

    return join_code(lines)


def _get_typeddict_class_def(name: str, model_type: TypedDictType) -> str:
    lines: t.List[str] = []

    if model_type is TypedDictType.PARAMS:
        lines.append(f'class {PARAMS_DICT}(te.TypedDict):')
    elif model_type is TypedDictType.DATA:
        lines.append(f'class {INPUT_DICT}(te.TypedDict):')

    lines.append('')

    return join_code(lines)


_LEXICON_TYPE_TO_PRIMITIVE_TYPEHINT = {
    models.LexString: 'str',
    models.LexInteger: 'int',
    models.LexBoolean: 'bool',
}


def _get_optional_typehint(type_hint: str, *, optional: bool) -> str:
    if optional:
        return f't.Optional[{type_hint}]'
    return type_hint


def _get_ref_typehint(nsid: NSID, field_type_def: models.LexRef, *, optional: bool) -> str:
    model_path, _ = _resolve_nsid_ref(nsid, field_type_def.ref)
    return _get_optional_typehint(f"'{model_path}'", optional=optional)


def _get_ref_union_typehint(nsid: NSID, field_type_def: models.LexRefUnion, *, optional: bool) -> str:
    def_names = []
    for ref in field_type_def.refs:
        import_path, _ = _resolve_nsid_ref(nsid, ref)
        def_names.append(import_path)

    # unbelievable but it's true. If schema doesn't describe the right type in Union
    # we should fall back to the plain data
    # maybe it's for the records that have custom fields... idk
    # ref: https://github.com/bluesky-social/atproto/blob/b01e47b61730d05a780f7a42667b91ccaa192e8e/packages/lex-cli/src/codegen/lex-gen.ts#L325
    # grep by "{$type: string; [k: string]: unknown}" string
    # TODO(MarshalX): use 'base.UnknownDict' and convert to DotDict
    # append 't.Dict[str, t.Any]' to def_names  # FIXME(MarshalX): support pydantic

    def_names = ', '.join([f"'{name}'" for name in def_names])
    def_field_meta = 'Field(default=None, discriminator="py_type")' if optional else 'Field(discriminator="py_type")'

    annotated_union = f'te.Annotated[t.Union[{def_names}], {def_field_meta}]'
    return _get_optional_typehint(annotated_union, optional=optional)


def _get_model_field_typehint(
    nsid: NSID,
    field_type_def: t.Union[models.LexPrimitive, models.LexArray, models.LexBlob],
    *,
    optional: bool,
    is_input_type: bool = False,
) -> str:
    field_type = type(field_type_def)

    if field_type == models.LexUnknown:
        # unknown type is a generic response with records or any not described type in the lexicon. for example, didDoc
        if is_input_type:
            return _get_optional_typehint("'UnknownInputType'", optional=optional)
        return _get_optional_typehint("'UnknownType'", optional=optional)

    type_hint = _LEXICON_TYPE_TO_PRIMITIVE_TYPEHINT.get(field_type)
    if type_hint:
        return _get_optional_typehint(type_hint, optional=optional)

    if field_type is models.LexArray:
        items_type_hint = _get_model_field_typehint(
            nsid, field_type_def.items, optional=False, is_input_type=is_input_type
        )
        return _get_optional_typehint(f't.List[{items_type_hint}]', optional=optional)

    if field_type is models.LexRef:
        return _get_ref_typehint(nsid, field_type_def, optional=optional)

    if field_type is models.LexRefUnion:
        return _get_ref_union_typehint(nsid, field_type_def, optional=optional)

    if field_type is models.LexCidLink:
        return _get_optional_typehint("'CIDType'", optional=optional)

    if field_type is models.LexBytes:
        # CAR file containing relevant blocks
        return _get_optional_typehint('t.Union[str, bytes]', optional=optional)

    if field_type is models.LexBlob:
        # yes, it returns blob,but actually it's blob ref here
        return _get_optional_typehint("'BlobRef'", optional=optional)

    raise ValueError(f'Unknown field type {field_type.__name__}')


def _get_model_field_value(  # noqa: C901
    field_type_def: t.Union[models.LexPrimitive, models.LexArray, models.LexBlob],
    alias_name: t.Optional[str] = None,
    *,
    optional: bool,
) -> str:
    not_set = object()

    default: t.Any = not_set
    alias: t.Union[str, not_set] = alias_name or not_set
    min_length: t.Union[int, not_set] = not_set
    max_length: t.Union[int, not_set] = not_set
    ge: t.Union[int, not_set] = not_set
    le: t.Union[int, not_set] = not_set
    frozen: t.Union[bool, not_set] = not_set

    field_type = type(field_type_def)

    if field_type == models.LexInteger:
        if field_type_def.default is not None:
            default = field_type_def.default
        if field_type_def.minimum is not None:
            ge = field_type_def.minimum
        if field_type_def.maximum is not None:
            le = field_type_def.maximum
        if field_type_def.const is not None:
            frozen = field_type_def.const

    elif field_type == models.LexString:
        if field_type_def.default is not None:
            default = field_type_def.default
        if field_type_def.const is not None:
            frozen = field_type_def.const
        if field_type_def.min_length is not None:
            min_length = field_type_def.min_length
        if field_type_def.max_length is not None:
            max_length = field_type_def.max_length
        # TODO (MarshalX): support knownValue, format, enum?

    elif field_type == models.LexBoolean:
        if field_type_def.default is not None:
            default = field_type_def.default
        if field_type_def.const is not None:
            frozen = field_type_def.const

    elif field_type is models.LexArray:
        if field_type_def.min_length is not None:
            min_length = field_type_def.min_length
        if field_type_def.max_length is not None:
            max_length = field_type_def.max_length

    if default is not_set and optional:
        default = None

    field_params = {
        'default': default,
        'alias': alias,
        'min_length': min_length,
        'max_length': max_length,
        'ge': ge,
        'le': le,
        'frozen': frozen,
    }

    values = []
    only_default = default is not not_set
    for name, value in field_params.items():
        if value is not_set:
            continue

        if name != 'default':
            only_default = False

        value_str = f'{name}='
        if isinstance(value, str):
            value_str += f"'{value}'"
        elif isinstance(value, bool):
            value_str += 'True' if value else 'False'
        else:
            value_str += str(value)

        values.append(value_str)

    if only_default:
        return 'None'

    if not values:
        return ''

    field_params_str = ', '.join(values)
    return f'Field({field_params_str})'


def _get_req_fields_set(lex_obj: t.Union[models.LexObject, models.LexXrpcParameters]) -> set:
    required_fields = set()
    if lex_obj.required:
        required_fields = set(lex_obj.required)

    if hasattr(lex_obj, 'nullable') and lex_obj.nullable:
        # TODO(MarshalX): not 100% the same thing. think about it
        required_fields = required_fields.difference(set(lex_obj.nullable))

    return required_fields


def _add_dot_to_end_if_not_exist(description: str) -> str:
    if description[-1] not in {'.', '?', '!'}:
        description += '.'
    return description


def _get_field_docstring(
    field_name: str, field_type: t.Union[models.LexPrimitive, models.LexArray, models.LexBlob]
) -> str:
    field_desc = field_type.description
    if field_desc is None:
        field_desc = gen_description_by_camel_case_name(field_name)

    return _add_dot_to_end_if_not_exist(field_desc)


def _get_model_docstring(
    nsid: t.Union[str, NSID],
    lex_object: t.Union[models.LexXrpcQuery, models.LexSubscription, models.LexObject, models.LexXrpcParameters],
    model_type: ModelType,
) -> str:
    model_desc = f' {lex_object.description}' if lex_object.description else ''
    model_desc = f'{model_type.value} model for :obj:`{nsid}`.{model_desc}'
    model_desc = _add_dot_to_end_if_not_exist(model_desc)

    doc_string = [f'{_(1)}"""{model_desc}"""', '']

    return join_code(doc_string)


@lru_cache(maxsize=None)
def _get_pydantic_reserved_names() -> t.Set[str]:
    from pydantic import BaseModel

    class Model(BaseModel):
        pass

    instance = Model()

    return {name for name in dir(instance) if not name.startswith('_')}


def _is_reserved_pydantic_name(name: str) -> bool:
    return name in _get_pydantic_reserved_names()


def _get_model(
    nsid: NSID, lex_object: t.Union[models.LexObject, models.LexXrpcParameters], *, is_input_type: bool = False
) -> str:
    required_fields = _get_req_fields_set(lex_object)

    fields = []
    optional_fields = []

    for field_name, field_type_def in lex_object.properties.items():
        is_optional = field_name not in required_fields

        alias_name = None
        snake_cased_field_name = convert_camel_case_to_snake_case(field_name)

        if snake_cased_field_name != field_name:
            # make aliases for fields with camel case names
            alias_name = field_name

        if _is_reserved_pydantic_name(snake_cased_field_name):
            # make aliases for fields with reserved names
            snake_cased_field_name += '_'  # add underscore to the end
            alias_name = field_name

        type_hint = _get_model_field_typehint(nsid, field_type_def, optional=is_optional, is_input_type=is_input_type)
        value = _get_model_field_value(field_type_def, alias_name, optional=is_optional)
        description = _get_field_docstring(field_name, field_type_def)

        value_def = f' = {value}' if value else ''
        field_def = f'{_(1)}{snake_cased_field_name}: {type_hint}{value_def} #: {description}'

        if is_optional:
            optional_fields.append(field_def)
        else:
            fields.append(field_def)

    optional_fields.sort()
    fields.sort()

    fields.extend(optional_fields)

    fields.append('')

    return join_code(fields)


def _get_typeddict(
    nsid: NSID, lex_object: t.Union[models.LexObject, models.LexXrpcParameters], *, is_input_type: bool = False
) -> str:
    required_fields = _get_req_fields_set(lex_object)

    fields: t.List[str] = []
    optional_fields: t.List[str] = []

    for field_name, field_type_def in lex_object.properties.items():
        is_optional = field_name not in required_fields

        snake_cased_field_name = convert_camel_case_to_snake_case(field_name)

        type_hint = _get_model_field_typehint(nsid, field_type_def, optional=is_optional, is_input_type=is_input_type)
        description = _get_field_docstring(field_name, field_type_def)

        # Allow optional params to actually be omitted from the dict entirely
        type_hint_defaulting = f'te.NotRequired[{type_hint}]' if is_optional else type_hint
        field_def = f'{_(1)}{snake_cased_field_name}: {type_hint_defaulting} #: {description}'

        if is_optional:
            optional_fields.append(field_def)
        else:
            fields.append(field_def)

    optional_fields.sort()
    fields.sort()

    fields.extend(optional_fields)

    if len(fields) == 0:
        fields.append(f'{_(1)}pass')

    fields.append('')

    return join_code(fields)


def _get_model_raw_data(name: str) -> str:
    lines = [f'#: {name} raw data type.', f'{name}: te.TypeAlias = bytes\n\n']
    return join_code(lines)


def _generate_params_model(nsid: NSID, definition: t.Union[models.LexXrpcQuery, models.LexSubscription]) -> str:
    lines = [_get_model_class_def(nsid.name, ModelType.PARAMS)]

    if definition.parameters:
        lines.append(_get_model_docstring(nsid, definition.parameters, ModelType.PARAMS))
        lines.append(_get_model(nsid, definition.parameters, is_input_type=True))

    lines.append(_get_typeddict_class_def(nsid.name, TypedDictType.PARAMS))

    if definition.parameters:
        lines.append(_get_typeddict(nsid, definition.parameters, is_input_type=True))

    return join_code(lines)


def _generate_xrpc_body_model(nsid: NSID, body: models.LexXrpcBody, model_type: ModelType) -> str:
    lines = []
    if body.schema_:
        if isinstance(body.schema_, models.LexObject):
            lines.append(_get_model_class_def(nsid.name, model_type))
            lines.append(_get_model_docstring(nsid, body.schema_, model_type))
            lines.append(_get_model(nsid, body.schema_, is_input_type=(model_type is ModelType.DATA)))
    else:
        if model_type is ModelType.DATA:
            model_name = INPUT_MODEL
        elif model_type is ModelType.RESPONSE:
            model_name = OUTPUT_MODEL
        else:
            raise ValueError('Wrong type or not implemented')

        lines.append(_get_model_raw_data(model_name))

    return join_code(lines)


def _generate_data_typedict(nsid: NSID, body: models.LexXrpcBody) -> str:
    lines: t.List[str] = []
    if isinstance(body.schema_, models.LexObject):
        lines.append(_get_typeddict_class_def(nsid.name, TypedDictType.DATA))
        lines.append(_get_typeddict(nsid, body.schema_, is_input_type=True))
    return join_code(lines)


def _generate_data_model(nsid: NSID, input_body: models.LexXrpcBody) -> str:
    return join_code(
        [_generate_xrpc_body_model(nsid, input_body, ModelType.DATA), _generate_data_typedict(nsid, input_body)]
    )


def _generate_response_model(nsid: NSID, output_body: models.LexXrpcBody) -> str:
    return _generate_xrpc_body_model(nsid, output_body, ModelType.RESPONSE)


_MAIN_MODEL_BACKWARD_COMPATIBILITY = """
class Main(Record):
    def __init_subclass__(cls, **data: t.Any) -> None:
        import warnings
        warnings.warn('Main class is deprecated. Use Record class instead.', DeprecationWarning, stacklevel=2)
        super().__init_subclass__(**data)

    def __init__(self, **data: t.Any) -> None:
        import warnings
        warnings.warn('Main class is deprecated. Use Record class instead.', DeprecationWarning, stacklevel=2)
        super().__init__(**data)
"""


def _generate_def_model(nsid: NSID, def_name: str, def_model: models.LexObject, model_type: ModelType) -> str:
    lines = [
        _get_model_class_def(def_name, model_type),
        _get_model_docstring(nsid, def_model, model_type),
        _get_model(nsid, def_model),
    ]

    def_type = f'{nsid}#{def_name}'
    if def_name == 'main':
        def_type = str(nsid)

    lines.append(f"{_(1)}py_type: te.Literal['{def_type}'] = Field(default='{def_type}', alias='$type', frozen=True)")
    lines.append('')

    # TODO(MarshalX): remove it
    if model_type is ModelType.RECORD:
        lines.append(_MAIN_MODEL_BACKWARD_COMPATIBILITY)

    return join_code(lines)


def _generate_def_token(nsid: NSID, def_name: str, def_model: models.LexToken) -> str:
    description = def_model.description
    if not description:
        description = gen_description_by_camel_case_name(def_name)

    lines = [
        f"{get_def_model_name(def_name)} = te.Literal['{nsid}#{def_name}'] #: {description}",
        '',
        '',
    ]
    return join_code(lines)


def _generate_def_array(nsid: NSID, def_name: str, def_model: models.LexArray) -> str:
    return f'{get_def_model_name(def_name)} = {_get_model_field_typehint(nsid, def_model, optional=False)}\n'


def _generate_def_string(nsid: NSID, def_name: str, def_model: models.LexString) -> str:
    # FIXME(MarshalX): support more fields. only known_values field is supported for now

    if not def_model.known_values:
        return ''

    union_types = []
    for known_value in def_model.known_values:
        if '#' in known_value:
            # reference to literal (token)
            model_path, _ = _resolve_nsid_ref(nsid, known_value)
            type_ = f"'{model_path}'"
        else:
            # literal
            # FIXME(MarshalX): not used for now
            type_ = f"te.Literal['{known_value}']"

        union_types.append(type_)

    final_type = f"t.Union[{', '.join(union_types)}]"

    description = def_model.description
    if not description:
        description = gen_description_by_camel_case_name(def_name)

    lines = [
        f'{get_def_model_name(def_name)} = {final_type} #: {description}',
        '',
        '',
    ]

    return join_code(lines)


def _generate_params_models(lex_db: builder.BuiltParamsModels) -> None:
    for nsid, defs in lex_db.items():
        _save_code_import_if_not_exist(nsid)

        definition = defs['main']
        if definition.parameters:
            if isinstance(definition.parameters, models.LexXrpcParameters):
                save_code_part(nsid, _generate_params_model(nsid, definition))
            else:
                # LexXrpcProcedure has parameters using another model
                raise ValueError('Wrong parameters type or not implemented')


def _generate_data_models(lex_db: builder.BuiltDataModels) -> None:
    for nsid, defs in lex_db.items():
        _save_code_import_if_not_exist(nsid)

        definition = defs['main']
        if definition.input:
            save_code_part(nsid, _generate_data_model(nsid, definition.input))


def _generate_response_models(lex_db: builder.BuiltResponseModels) -> None:
    for nsid, defs in lex_db.items():
        _save_code_import_if_not_exist(nsid)

        definition = defs['main']
        if definition.output:
            save_code_part(nsid, _generate_response_model(nsid, definition.output))


def _generate_def_models(lex_db: builder.BuiltDefModels) -> None:
    for nsid, defs in lex_db.items():
        _save_code_import_if_not_exist(nsid)

        for def_name, def_model in defs.items():
            if isinstance(def_model, models.LexToken):
                save_code_part(nsid, _generate_def_token(nsid, def_name, def_model))
            elif isinstance(def_model, models.LexString):
                save_code_part(nsid, _generate_def_string(nsid, def_name, def_model))
            elif isinstance(def_model, models.LexObject):
                save_code_part(nsid, _generate_def_model(nsid, def_name, def_model, ModelType.DEF))
            elif isinstance(def_model, models.LexArray):
                save_code_part(nsid, _generate_def_array(nsid, def_name, def_model))
            else:
                raise ValueError(f'Unhandled type {type(def_model)}')


def _generate_record_sugar_models(nsid: NSID) -> str:
    lines = []
    for model in [
        RECORD_GET_RESPONSE_MODEL_TEMPLATE,
        RECORD_LIST_RESPONSE_MODEL_TEMPLATE,
        RECORD_CREATE_RESPONSE_MODEL_TEMPLATE,
    ]:
        lines.append(model.format(record_import=get_import_path(nsid)))

    return join_code(lines)


def _generate_record_models(lex_db: builder.BuiltRecordModels) -> None:
    for nsid, defs in lex_db.items():
        _save_code_import_if_not_exist(nsid)

        for def_name, def_record in defs.items():
            if isinstance(def_record, models.LexRecord):
                # TODO(MarshalX): Process somehow def_record.key?
                save_code_part(nsid, _generate_def_model(nsid, def_name, def_record.record, ModelType.RECORD))
                save_code_part(nsid, _generate_record_sugar_models(nsid))


def _generate_record_type_database(lex_db: builder.BuiltRecordModels) -> None:
    type_conversion_lines = ['from atproto_client import models', 'RECORD_TYPE_TO_MODEL_CLASS = {']

    import_lines = [
        'import typing as t',
        'import typing_extensions as te',
        'from pydantic import Field',
        'if t.TYPE_CHECKING:',
        f'{_(4)}from atproto_client.models import base',
        f'{_(4)}from atproto_client import models',
        f'{_(4)}from atproto_client.models import dot_dict',
        '',
    ]
    unknown_record_type_hint_lines = ['UnknownRecordType: te.TypeAlias = t.Union[']
    unknown_record_type_pydantic_lines = ['UnknownRecordTypePydantic = te.Annotated[t.Union[']

    for nsid, defs in lex_db.items():
        _save_code_import_if_not_exist(nsid)

        for def_name, def_record in defs.items():
            # for now, there are records only in under "main" definition name.
            # need to rework a bit if this behavior is changed
            if isinstance(def_record, models.LexRecord):
                class_name = get_record_model_name(def_name)
                record_type = str(nsid)

                path_to_class = f'models.{get_import_path(nsid)}.{class_name}'

                type_conversion_lines.append(f"'{record_type}': {path_to_class},")

                unknown_record_type_hint_lines.append(f"{_(4)}'{path_to_class}',")
                unknown_record_type_pydantic_lines.append(f"{_(4)}'{path_to_class}',")

    type_conversion_lines.append('}')

    unknown_record_type_hint_lines.append(']')
    unknown_record_type_pydantic_lines.append('], Field(discriminator="py_type")]')
    unknown_record_type_pydantic_lines.append(
        "UnknownType: te.TypeAlias = t.Union[UnknownRecordTypePydantic, 'dot_dict.DotDictType']"
    )
    unknown_record_type_pydantic_lines.append(
        'UnknownInputType: te.TypeAlias = t.Union[UnknownType, t.Dict[str, t.Any]]'
    )
    unknown_type_lines = [*import_lines, *unknown_record_type_hint_lines, *unknown_record_type_pydantic_lines]

    write_code(_MODELS_OUTPUT_DIR.joinpath('type_conversion.py'), join_code(type_conversion_lines))
    write_code(_MODELS_OUTPUT_DIR.joinpath('unknown_type.py'), join_code(unknown_type_lines))


def _generate_init_files(root_package_path: Path) -> None:
    # One of the ways that I tried. Doesn't work well due to circular imports
    for root, dirs, files in sorted(os.walk(root_package_path)):
        root_path = Path(root)

        import_lines = []
        for dir_name in sorted(dirs):
            if dir_name.startswith('__'):
                continue

            import_parts = root_path.parts[root_path.joinpath(dir_name).parts.index(_MODELS_OUTPUT_DIR.parent.name) :]
            from_import = '.'.join(import_parts)

            if dir_name in {'app', 'com'}:
                continue

            import_lines.append(f'from {from_import} import {dir_name}')

        for file_name in sorted(files):
            if file_name.startswith('__'):
                continue

            import_parts = root_path.parts[root_path.parts.index(_MODELS_OUTPUT_DIR.parent.name) :]
            from_import = '.'.join(import_parts)

            import_lines.append(f'from atproto_client.{from_import} import {file_name[:-3]}')

        if root_path.name == 'models':
            # FIXME skip for now. should be generated too
            continue

        if root_path.name == '__pycache__':
            continue

        write_code(root_path.joinpath('__init__.py'), join_code(import_lines))


def _generate_empty_init_files(root_package_path: Path) -> None:
    for root, dirs, files in os.walk(root_package_path):
        root_path = Path(root)

        for dir_name in dirs:
            if dir_name.startswith('__'):
                continue

            if dir_name in {'app', 'com'}:
                continue

        for file_name in files:
            if file_name.startswith('__'):
                continue

        if root_path.name == 'models':
            # FIXME skip for now. should be generated too
            continue

        if root_path.name == '__pycache__':
            continue

        write_code(root_path.joinpath('__init__.py'), DISCLAIMER)


def _generate_import_aliases(root_package_path: Path) -> None:
    # is generates __init__.py file if models dir with aliases like this:
    # from atproto_client.models.app.bsky.actor import defs as AppBskyActorDefs # noqa: ERA001

    import_lines = []
    ids_db = ['class _Ids:']
    for root, __, files in sorted(os.walk(root_package_path)):
        root_path = Path(root)

        if root_path == root_package_path:
            continue

        for file in sorted(files):
            if file.startswith(('.', '__', '.pyc')):
                continue
            if '.cpython-' in file:
                continue

            import_parts = root_path.parts[root_path.parts.index(_MODELS_OUTPUT_DIR.parent.name) :]
            from_import = '.'.join(import_parts)

            nsid_parts = list(root_path.parts[root_path.parts.index('models') + 1 :])
            method_name_parts = file[:-3].split('_')
            alias_name = ''.join([p.capitalize() for p in [*nsid_parts, *method_name_parts]])

            camel_case_method_name = method_name_parts[0] + ''.join(ele.title() for ele in method_name_parts[1:])
            method_path = f"{'.'.join(nsid_parts)}.{camel_case_method_name}"
            ids_db.append(f"{_(1)}{alias_name}: str = '{method_path}'")

            import_lines.append(f'from {from_import} import {file[:-3]} as {alias_name}')

    import_lines.append(
        'from atproto_client.models.utils import '
        'create_strong_ref, '
        'get_model_as_dict, '
        'get_model_as_json, '
        'get_or_create, '
        'is_record_type'
    )
    import_lines.append('from atproto_client.models.models_loader import load_models')

    ids_db.append('ids = _Ids()')
    import_lines.extend(ids_db)

    import_lines.append('load_models()')

    write_code(_MODELS_OUTPUT_DIR.joinpath('__init__.py'), join_code(import_lines))


def generate_models(lexicon_dir: t.Optional[Path] = None, output_dir: t.Optional[Path] = None) -> None:
    if lexicon_dir:
        builder.lexicon_dir.set(lexicon_dir)

    if output_dir:
        # TODO(MarshalX): Temp hack for CLI. Pass output_dir everywhere.
        global _MODELS_OUTPUT_DIR
        _MODELS_OUTPUT_DIR = output_dir

    _generate_params_models(builder.build_params_models())
    _generate_data_models(builder.build_data_models())
    _generate_response_models(builder.build_response_models())
    _generate_def_models(builder.build_def_models())

    _generate_record_models(builder.build_record_models())
    _generate_record_type_database(builder.build_record_models())

    _generate_empty_init_files(_MODELS_OUTPUT_DIR)
    _generate_import_aliases(_MODELS_OUTPUT_DIR)

    format_code(_MODELS_OUTPUT_DIR)
